/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Virtual server
*  SCOPE :              Virtual server
*  FILE :               virtualdbserver.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#ifndef __virtualdbserver_HXX
#include "virtualdbserver.hxx"
#endif

#ifndef __dbvue_HXX
#include "dbvue.hxx"
#endif

#ifndef __callbackclient_clnt_HXX
#include "callbackclient_clnt.hxx"
#endif

#ifndef __var_H
#include "var.h"
#endif

#ifndef __dbserver_HXX
#include "dbserver.hxx"
#endif

#ifndef __yooexcept_HXX
#include "yooexcept.hxx"
#endif

#ifndef __volume_HXX
#include "volume.hxx"
#endif

#ifndef __yoofunc_HXX
#include "yoofunc.hxx"
#endif

#ifndef __cluster_HXX
#include "cluster.hxx"
#endif

#ifndef __oid_HXX
#include "oid.hxx"
#endif

#ifndef __pagehdlmem_HXX
#include "pagehdlmem.hxx"
#endif

#ifndef __virtualmemory_HXX
#include "virtualmemory.hxx"
#endif

/*****************************************************************************/


/*= EXTERNAL FUNCTION =======================================================*/
VirtualDBServer::VirtualDBServer(const char* base)
:_delta(FALSE),_name(strdup(base))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  for (int i = 0; i < MAX_VOL; i++) {
    _tabvue[i] = NULL ;
    _tabname[i] = NULL ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
VirtualDBServer::~VirtualDBServer() 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  for (int i = 0; i < MAX_VOL; i++) {
    if (_tabname[i] != NULL) delete _tabname[i] ;
  }
  delete _name ;
}


/*= EXTERNAL FUNCTION =======================================================*/
DBVue* VirtualDBServer::server(int vol)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_tabvue[vol] == NULL) return NULL ;
  if (_tabvue[vol]->isConnected() == FALSE) {
    this->_vues.insertRef(_tabvue[vol], 0) ;
    _tabvue[vol]->connect(_tabname[vol], _delta) ;
  }
  return _tabvue[vol] ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBServer::createVue(const char* server, ListPtr<Volume>* vol)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBVue* vue ;
  if ( (YOO_DBServer != NULL) &&       
       ( (strcmp(MONOUSER,YOO_DBServer->name())==0) ||
	 (strcmp(server,YOO_DBServer->name())==0) ) ) { 
    vue = new LocalDBVue() ;
    this->_vues.insertRef(vue, 0) ;
  } else {
    vue = new ProxyDBVue ;
  }
  IterListPtr<Volume> iter(*vol) ;
  for (; iter.more(); iter.next()) {
    _tabvue[iter.get()->volid()] = vue ;
    _tabname[iter.get()->volid()] = strdup(server) ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::lock(PID& pid, Lock l)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBVue* vue ;
  vue = this->server(pid.volNo) ;
  if (vue == NULL) call_throw(YOO_ERR_READ, "Invalid Address") ;
  return vue->lock(pid, l) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBServer::readfirst(Page* p, PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBVue* vue ;
  vue = this->server(pid.volNo) ;
  if (vue == NULL) call_throw(YOO_ERR_READ, "Invalid Address") ;
  vue->readfirst(*p, pid) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBServer::readnext(Page* p, PID& pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBVue* vue ;
  vue = this->server(pid.volNo) ;
  if (vue == NULL) call_throw(YOO_ERR_READ, "Invalid Address") ;
  vue->readnext(*p, pid) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
PageHdlMem* VirtualDBServer::allocation(Cluster* clu, int num)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  DBVue* vue ;
  vue = this->server(clu->getVol()) ;
  if (vue == NULL) call_throw(YOO_ERR_COM, "Trouble with server") ;
  PID pid = vue->allocation(clu->getVol(), num) ;

  if (pid == NULL_PID) {
    yoo_error(YOO_ERR_VOL, "Volume unable to alloc") ;
    yoo_exit(YOO_ERR_VOL) ;
  }

  PageHdlMem* pd = YOO_VirtualMemory->allocation(pid, num, YOO_EXCL) ;
  pd->page()->init(clu, pid, num) ;
  return pd ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Page* VirtualDBServer::findPage(PID pid)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (pid == NULL_PID) return NULL ;
  DBVue* vue ;  
  vue = this->server(pid.volNo) ;
  if (vue == NULL) call_throw(YOO_ERR_READ, "Invalid Address") ;

  PageHdlMem* pd ;
  if ((pd = (PageHdlMem*)_keep[pid]) != NULL) {
    if ((YOO_Callback != NULL) && (YOO_Callback->askLock(pid))) {
      YOO_VirtualMemory->refresh(pd) ;
      pd->page()->receive() ;      
      return pd->page() ;
    }
    _keep -= pid ;
  }

  pid.setsize(yoo_pageSize(pid.size()*SLOT_SIZE + PAGE_HEADER_SIZE)) ;
  pd = YOO_VirtualMemory->allocation(pid,pid.size(),YOO_SHARE);
  if (pid.size() == 1) pd->insertCluster() ;
  return pd->page() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::checkpoint()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterListPtr<PageHdlMem> iter(_touchPage) ; 
  for (; iter.more(); iter.next()) {
    DBVue* vue ;
    PID pid = iter.get()->pid() ;
    vue = this->server(pid.volNo) ;
    if (vue == NULL) call_throw(YOO_ERR_COM, "Trouble with server") ;
    if (iter.get()->isdeleted()) {
      vue->freepage(pid) ;
    } else {
      vue->write(*iter.get()->page(), pid) ;
    }
    if (vue->svcerror()) return -1 ;
  }
  _touchPage.clear() ;
  this->sync() ;
  return 0 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::upgrade(PageHdlMem** pd_t)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  ListVal<PID> eject_list ;
  for (int i = 0; i < YOO_VirtualMemory->size(); i++) {
    if (pd_t[i] == NULL) continue ;
    int volno = pd_t[i]->pid().volNo ;
    _tabvue[volno]->pidlist() += pd_t[i]->pid() ;
  }
  IterListVal<DBVue> iter(_vues) ;
  for (; iter.more(); iter.next()) {
    if (iter.get().pidlist().count() != 0) {
      iter.get().upgrade(iter.get().pidlist(), eject_list) ;
      if (iter.get().svcerror()) return -1 ;      
      iter.get().pidlist().clear() ;
    }
  }
  IterListVal<PID> iter2(eject_list) ;
  for (; iter2.more(); iter2.next()) {
    YOO_VirtualMemory->findPageHdl(iter2.get())->eject() ;
  }
  return 0 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::release(PageHdlMem** pd_t)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  ListVal<PID> keep_list ;
  ListVal<PID> eject_list ;
  for (int i = 0; i < YOO_VirtualMemory->size(); i++) {
    if (pd_t[i] == NULL) continue ;
    if (pd_t[i]->iskept()) {
      pd_t[i]->eject() ;
      continue ;
    }
    int volno = pd_t[i]->pid().volNo ;
    if (pd_t[i]->shortpage()) { _tabvue[volno]->pidlist()+=pd_t[i]->pid() ; }
  }
  IterListVal<DBVue> iter(_vues) ;
  for (; iter.more(); iter.next()) {
    if (iter.get().pidlist().count() != 0) {
      iter.get().release(_delta, iter.get().pidlist(), keep_list, eject_list) ;
      if (iter.get().svcerror()) return -1 ;      
      iter.get().pidlist().clear() ;
    }
  }
  IterListVal<PID> iter2(keep_list) ;
  for (; iter2.more(); iter2.next()) {
    PageHdlMem* pd = YOO_VirtualMemory->findPageHdl(iter2.get()) ;
    _keep += pd ;    
    pd->kept() ; 
  }
  if (YOO_Callback != NULL) YOO_Callback->keepLock(keep_list) ;
  IterListVal<PID> iter3(eject_list) ;
  for (; iter3.more(); iter3.next()) {
    YOO_VirtualMemory->eject( YOO_VirtualMemory->findPageHdl(iter3.get()) ) ;
  }
  return 0 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::abort()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterListVal<DBVue> iter(_vues) ;
  for (; iter.more(); iter.next()) {
    iter.get().abort(_delta) ;
    if (iter.get().svcerror()) {
      yoo_error(YOO_ERR_COM, "Problem with server") ;
      yoo_exit(YOO_ERR_COM) ;
    }
  }
  _touchPage.clear() ;
  return 0 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::secondphase()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterListVal<DBVue> iter(_vues) ;
  for (; iter.more(); iter.next()) {
    if ((iter.get().secondphase() == -1) || (iter.get().svcerror())) {
      yoo_error(YOO_ERR_COM, "Problem with server") ;
      yoo_exit(YOO_ERR_COM) ;
    }
  }
  return 0 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBServer::sync()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterListVal<DBVue> iter(_vues) ;
  for (; iter.more(); iter.next()) {
    if ((iter.get().sync() == -1) || (iter.get().svcerror())) {
      yoo_error(YOO_ERR_COM, "Problem with server") ;
      yoo_exit(YOO_ERR_COM) ;
    }
  }
  return 0 ;
}


